% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% %                                                                       %
% %    Matlab script for lecture                                          %
% %    "Simple differential equations - introduction using Matlab"        %
% %    by Christian Rummel                                                %
% %                                                                       %
% %    BENESCO lecture series                                             %
% %    "Signal Analysis and Brain Oscillations in Health and Disease"     %
% %    Bern, 2017/11/24                                                   %
% %                                                                       %
% %    Script demonstrates integration of the damped autonomous harmonic  %
% %    oscillator.                                                        %
% %    Euler's first order integrator is compared to fourth order Runge-  %
% %    Kutta.                                                             %
% %                                                                       %
% %    written by C. Rummel, 2017/11/20                                   %
% %                                                                       %
% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

clear all
close all
clc

% parameters of the integrator

d        =     2                       % dimension of time series
len      = 20000                       % length of time series
dt_range = 0.01*[1 2 5 10 20 50 100]   % range of integration step sizes

% define parameters for usage also in function for derivative 

global gamma sig omega0

gamma    = 0.05;
sig      = 1;
omega0   = 1;

% number of repetitions for error statistics

N_stat   = 10

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% loops over step sizes and repetitions

err_ave = zeros(2,length(dt_range));

for n = 1:length(dt_range)
    waitbar((n-1)/length(dt_range));

    dt      = dt_range(n);
    N       = round(len*dt_range(1)/dt);
    
    err_E   = [];
    err_RK4 = [];
    
    for n_stat = 1:N_stat

% sample random initial condition

        x_0        = randn(d,1);

% integrate damped oscillator twice

        [x_E, t]   = Euler       (@harmosc, d, N, dt, x_0);
        [x_RK4, t] = RungeKutta4 (@harmosc, d, N, dt, x_0);

% calculate exact solution

        A          = [0 1; -sig*omega0^2 -gamma];
        x_exact    = zeros(size(x_E));
        
        for k = 1:N
            x_exact(:,k) = expm(A*t(k)) * x_0;
        end

% display examples of integrated time series and exact solution

        if n_stat==1
            figure(1000+10*n+1)
            title('trajectory');
            hold on
            plot(x_E(1,:),x_E(2,:),'b');
            plot(x_RK4(1,:),x_RK4(2,:),'r');
            plot(x_exact(1,:),x_exact(2,:),'k');

            figure(1000+10*n+2)
            title('coordinate');
            hold on
            plot(t,x_E(1,:),'b');
            plot(t,x_RK4(1,:),'r');
            plot(t,x_exact(1,:),'k');
        
            figure(1000+10*n+3)
            title('velocity');
            hold on
            plot(t,x_E(2,:),'b');
            plot(t,x_RK4(2,:),'r');
            plot(t,x_exact(2,:),'k');
            
            pause
        end

% accumulate relative errors along trajectory

        err_E   = [err_E   rel_err(x_E,x_exact,t)];
        err_RK4 = [err_RK4 rel_err(x_RK4,x_exact,t)];
    end
     
% calculate mean relative errors
    
    err_ave(1,n) = nanmean(err_E);
    err_ave(2,n) = nanmean(err_RK4);
end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

% display error as a function of step size

figure(10)
hold on
plot(dt_range,err_ave(1,:),'b');
plot(dt_range,err_ave(2,:),'r');
set(gca,'XScale','log');
set(gca,'YScale','log');


